import { ComputeBudgetProgram, Keypair, SendTransactionError, TransactionExpiredBlockheightExceededError, } from '@solana/web3.js';
import { encoding } from '@wormhole-foundation/sdk-connect';
import { SolanaPlatform } from './platform.js';
// Add priority fee according to 90th percentile of recent fees paid
const DEFAULT_PRIORITY_FEE_PERCENTILE = 0.9;
// returns a SignOnlySigner for the Solana platform
export async function getSolanaSigner(rpc, privateKey) {
    const [_, chain] = await SolanaPlatform.chainFromRpc(rpc);
    return new SolanaSigner(chain, Keypair.fromSecretKey(encoding.b58.decode(privateKey)), rpc);
}
// returns a SignAndSendSigner for the Solana platform
export async function getSolanaSignAndSendSigner(rpc, privateKey, opts) {
    const [_, chain] = await SolanaPlatform.chainFromRpc(rpc);
    const kp = typeof privateKey === 'string'
        ? Keypair.fromSecretKey(encoding.b58.decode(privateKey))
        : privateKey;
    if (opts?.priorityFeePercentile && opts?.priorityFeePercentile > 1.0)
        throw new Error('priorityFeePercentile must be a number between 0 and 1');
    return new SolanaSendSigner(rpc, chain, kp, opts?.debug ?? false, opts?.priorityFeePercentile ?? 0, opts?.sendOpts);
}
export class SolanaSigner {
    _chain;
    _keypair;
    _rpc;
    _debug;
    constructor(_chain, _keypair, _rpc, _debug = false) {
        this._chain = _chain;
        this._keypair = _keypair;
        this._rpc = _rpc;
        this._debug = _debug;
    }
    chain() {
        return this._chain;
    }
    address() {
        return this._keypair.publicKey.toBase58();
    }
    async sign(tx) {
        const { blockhash } = await SolanaPlatform.latestBlock(this._rpc);
        const signed = [];
        for (const txn of tx) {
            const { description, transaction: { transaction, signers: extraSigners }, } = txn;
            console.log(`Signing: ${description} for ${this.address()}`);
            if (this._debug)
                logTxDetails(transaction);
            transaction.recentBlockhash = blockhash;
            transaction.partialSign(this._keypair, ...(extraSigners ?? []));
            signed.push(transaction.serialize());
        }
        return signed;
    }
}
export class SolanaSendSigner {
    _rpc;
    _chain;
    _keypair;
    _debug;
    _priorityFeePercentile;
    _sendOpts;
    constructor(_rpc, _chain, _keypair, _debug = false, _priorityFeePercentile = 0.0, _sendOpts) {
        this._rpc = _rpc;
        this._chain = _chain;
        this._keypair = _keypair;
        this._debug = _debug;
        this._priorityFeePercentile = _priorityFeePercentile;
        this._sendOpts = _sendOpts;
        this._sendOpts = this._sendOpts ?? {
            preflightCommitment: this._rpc.commitment,
        };
    }
    chain() {
        return this._chain;
    }
    address() {
        return this._keypair.publicKey.toBase58();
    }
    // Handles retrying a Transaction if the error is deemed to be
    // recoverable. Currently handles:
    // - Transaction expired
    // - Blockhash not found
    // - Not enough bytes (storage account not seen yet)
    retryable(e) {
        // Tx expired, set a new block hash and retry
        if (e instanceof TransactionExpiredBlockheightExceededError)
            return true;
        // Besides tx expiry, only handle SendTransactionError
        if (!(e instanceof SendTransactionError))
            return false;
        // Only handle simulation errors
        if (!e.message.includes('Transaction simulation failed'))
            return false;
        // Blockhash not found, similar to expired, resend with new blockhash
        if (e.message.includes('Blockhash not found'))
            return true;
        // Find the log message with the error details
        const loggedErr = e.logs?.find((log) => log.startsWith('Program log: Error: '));
        // who knows
        if (!loggedErr)
            return false;
        // Probably caused by storage account not seen yet
        if (loggedErr.includes('Not enough bytes'))
            return true;
        if (loggedErr.includes('Unexpected length of input'))
            return true;
        return false;
    }
    async signAndSend(tx) {
        let { blockhash, lastValidBlockHeight } = await SolanaPlatform.latestBlock(this._rpc);
        const txids = [];
        for (const txn of tx) {
            const { description, transaction: { transaction, signers: extraSigners }, } = txn;
            console.log(`Signing: ${description} for ${this.address()}`);
            if (this._priorityFeePercentile && this._priorityFeePercentile > 0)
                transaction.add(...(await createPriorityFeeInstructions(this._rpc, transaction, [], this._priorityFeePercentile)));
            if (this._debug)
                logTxDetails(transaction);
            // Try to send the transaction up to 5 times
            const maxRetries = 5;
            for (let i = 0; i < maxRetries; i++) {
                try {
                    transaction.recentBlockhash = blockhash;
                    transaction.partialSign(this._keypair, ...(extraSigners ?? []));
                    const txid = await this._rpc.sendRawTransaction(transaction.serialize(), this._sendOpts);
                    txids.push(txid);
                    break;
                }
                catch (e) {
                    // No point checking if retryable if we're on the last retry
                    if (i === maxRetries - 1)
                        throw e;
                    // If it's not retryable, throw
                    if (!this.retryable(e))
                        throw e;
                    // If it is retryable, we need to grab a new block hash
                    const { blockhash: newBlockhash, lastValidBlockHeight: newBlockHeight, } = await SolanaPlatform.latestBlock(this._rpc);
                    lastValidBlockHeight = newBlockHeight;
                    blockhash = newBlockhash;
                }
            }
        }
        // Wait for finalization
        const results = await Promise.all(txids.map((signature) => this._rpc.confirmTransaction({
            signature,
            blockhash,
            lastValidBlockHeight,
        }, this._rpc.commitment)));
        const erroredTxs = results
            .filter((result) => result.value.err)
            .map((result) => result.value.err);
        if (erroredTxs.length > 0)
            throw new Error(`Failed to confirm transaction: ${erroredTxs}`);
        return txids;
    }
}
export function logTxDetails(transaction) {
    console.log(transaction.signatures);
    console.log(transaction.feePayer);
    transaction.instructions.forEach((ix) => {
        console.log('Program', ix.programId.toBase58());
        console.log('Data: ', ix.data.toString('hex'));
        console.log('Keys: ', ix.keys.map((k) => [k, k.pubkey.toBase58()]));
    });
}
export async function createPriorityFeeInstructions(connection, transaction, lockedWritableAccounts = [], feePercentile = DEFAULT_PRIORITY_FEE_PERCENTILE, minPriorityFee = 0) {
    if (lockedWritableAccounts.length === 0) {
        lockedWritableAccounts = transaction.instructions
            .flatMap((ix) => ix.keys)
            .map((k) => (k.isWritable ? k.pubkey : null))
            .filter((k) => k !== null);
    }
    return await determineComputeBudget(connection, transaction, lockedWritableAccounts, feePercentile, minPriorityFee);
}
export async function determineComputeBudget(connection, transaction, lockedWritableAccounts = [], feePercentile = DEFAULT_PRIORITY_FEE_PERCENTILE, minPriorityFee = 0) {
    let computeBudget = 250_000;
    let priorityFee = 1;
    try {
        const simulateResponse = await connection.simulateTransaction(transaction);
        if (simulateResponse.value.err) {
            console.error(`Error simulating Solana transaction: ${simulateResponse.value.err}`);
        }
        if (simulateResponse?.value?.unitsConsumed) {
            // Set compute budget to 120% of the units used in the simulated transaction
            computeBudget = Math.round(simulateResponse.value.unitsConsumed * 1.2);
        }
    }
    catch (e) {
        console.error(`Failed to calculate compute unit limit for Solana transaction: ${e}`);
    }
    try {
        priorityFee = await determinePriorityFee(connection, lockedWritableAccounts, feePercentile);
    }
    catch (e) {
        console.error(`Failed to calculate compute unit price for Solana transaction: ${e}`);
        return [];
    }
    priorityFee = Math.max(priorityFee, minPriorityFee);
    return [
        ComputeBudgetProgram.setComputeUnitLimit({
            units: computeBudget,
        }),
        ComputeBudgetProgram.setComputeUnitPrice({
            microLamports: priorityFee,
        }),
    ];
}
async function determinePriorityFee(connection, lockedWritableAccounts = [], percentile) {
    // https://twitter.com/0xMert_/status/1768669928825962706
    let fee = 1; // Set fee to 1 microlamport by default
    try {
        const recentFeesResponse = await connection.getRecentPrioritizationFees({
            lockedWritableAccounts,
        });
        if (recentFeesResponse) {
            // Get 75th percentile fee paid in recent slots
            const recentFees = recentFeesResponse
                .map((dp) => dp.prioritizationFee)
                .filter((dp) => dp > 0)
                .sort((a, b) => a - b);
            if (recentFees.length > 0) {
                const medianFee = recentFees[Math.floor(recentFees.length * percentile)];
                fee = Math.max(fee, medianFee);
            }
        }
    }
    catch (e) {
        console.error('Error fetching Solana recent fees', e);
    }
    return fee;
}
//# sourceMappingURL=signer.js.map